import 'dart:async';
import 'package:flutter/services.dart';

import 'internal/chat_method_keys.dart';
import 'chat_client.dart';
import 'models/chat_error.dart';
import 'models/chat_userInfo.dart';

///
/// The user attribute manager class, which gets and sets the user attributes.
///
class ChatUserInfoManager {
  static const _channelPrefix = 'com.chat.im';
  static const MethodChannel _channel = const MethodChannel(
      '$_channelPrefix/chat_userInfo_manager', JSONMethodCodec());

  ChatUserInfo? _ownUserInfo;

  // The map of effective contacts.
  Map<String, ChatUserInfo> _effectiveUserInfoMap = Map();

  ///
  /// Modifies the user attributes of the current user.
  ///
  /// Param [userInfo] The user attributes to be modified.
  ///
  /// **Throws**  A description of the exception. See {@link ChatError}.
  ///
  Future<void> updateUserInfo(ChatUserInfo userInfo) async {
    Map req = {'userInfo': userInfo.toJson()};
    Map result =
        await _channel.invokeMethod(ChatMethodKeys.updateOwnUserInfo, req);
    try {
      ChatError.hasErrorFromResult(result);
    } on ChatError catch (e) {
      throw e;
    }
  }

  ///
  /// Gets the current user's attributes from the server.
  ///
  /// Param [expireTime] The time period(seconds) when the user attibutes in the cache expire. If the interval between two calles is less than or equal to the value you set in the parameter, user attributes are obtained directly from the local cache; otherwise, they are obtained from the server. For example, if you set this parameter to 120(2 minutes), once this method is called again within 2 minutes, the SDK returns the attributes obtained last time.
  ///
  /// **Return**  The user properties that are obtained. See {@link ChatUserInfo}.
  ///
  /// **Throws**  A description of the exception. See {@link ChatError}.
  ///
  Future<ChatUserInfo?> fetchOwnInfo({int expireTime = 0}) async {
    String? currentUser = await ChatClient.getInstance.getCurrentUsername();
    if (currentUser != null) {
      try {
        Map<String, ChatUserInfo> ret = await fetchUserInfoById(
          [currentUser],
          expireTime: expireTime,
        );
        _ownUserInfo = ret.values.first;
      } on ChatError catch (e) {
        throw e;
      }
    }
    return _ownUserInfo;
  }

  ///
  /// Gets user attributes of the specified users.
  ///
  /// Param [userIds] The username array.
  ///
  /// Param [expireTime] The time period(seconds) when the user attibutes in the cache expire. If the interval between two calles is less than or equal to the value you set in the parameter, user attributes are obtained directly from the local cache; otherwise, they are obtained from the server. For example, if you set this parameter to 120(2 minutes), once this method is called again within 2 minutes, the SDK returns the attributes obtained last time.
  ///
  /// **Return** A map that contains key-value pairs where the key is the user ID and the value is user attributes.
  ///
  /// **Throws**  A description of the exception. See {@link ChatError}.
  ///
  Future<Map<String, ChatUserInfo>> fetchUserInfoById(
    List<String> userIds, {
    int expireTime = 0,
  }) async {
    List<String> needReqIds = userIds
        .where((element) =>
            !_effectiveUserInfoMap.containsKey(element) ||
            (_effectiveUserInfoMap.containsKey(element) &&
                DateTime.now().millisecondsSinceEpoch -
                        _effectiveUserInfoMap[element]!.expireTime >
                    expireTime * 1000))
        .toList();
    Map<String, ChatUserInfo> resultMap = Map();

    userIds.forEach((element) {
      if (_effectiveUserInfoMap.containsKey(element)) {
        resultMap[element] = _effectiveUserInfoMap[element]!;
      }
    });
    if (needReqIds.length == 0) {
      return resultMap;
    }

    Map req = {'userIds': needReqIds};
    Map result =
        await _channel.invokeMethod(ChatMethodKeys.fetchUserInfoById, req);

    try {
      ChatError.hasErrorFromResult(result);
      result[ChatMethodKeys.fetchUserInfoById]?.forEach((key, value) {
        ChatUserInfo eUserInfo = ChatUserInfo.fromJson(value);
        resultMap[key] = eUserInfo;
        _effectiveUserInfoMap[key] = eUserInfo;
      });
      return resultMap;
    } on ChatError catch (e) {
      throw e;
    }
  }

  @Deprecated("Switch to using ChatUserInfoManager#updateUserInfo instead.")
  Future<void> updateOwnerUserInfo(ChatUserInfo userInfo) async {
    Map req = {'userInfo': userInfo.toJson()};
    Map result =
        await _channel.invokeMethod(ChatMethodKeys.updateOwnUserInfo, req);
    try {
      ChatError.hasErrorFromResult(result);
    } on ChatError catch (e) {
      throw e;
    }
  }

  void clearUserInfoCache() {
    _ownUserInfo = null;
    _effectiveUserInfoMap.clear();
  }
}
